bits 16

%define ResetScreen       +15 bytes
%define StepCounter       +29 bytes

cpu 8086
org 0x7c00

;; Constants

;; Boot sector size in bytes
%assign BootSector.Size 512

;; Words in 16 bit x86 are 2 bytes
%assign WordSize 2

%assign DigitsNumber 6

;; This is the value to store in segment register to access the MDA and the VGA text buffer.
;; In 16 bit x86, segmented memory accesses are of the form:
;;
;;   (segment register) * 0x10 + (offset register)
;;
;; The MDA text buffer is at 0xb00000, so if 0xb000 is stored in a segment
;; register, then memory access instructions will be relative to the MDA text
;; buffer, allowing easier access. For example, trying to access the nth byte of
;; memory will *actually* access the nth byte of the text buffer.
%assign TextBuf.Seg           0xb000
%assign Vga.TextBuf.Offset    0x8000

%assign BitMap 0x8000

;; Dimensions of text buffer
%assign TextBuf.Width  80
%assign TextBuf.Height 25
%assign TextBuf.Size (TextBuf.Width * TextBuf.Height)

;; Macro to get the index of a text buffer cell from coordinates
%define TextBuf.Index(y, x) ((y) * TextBuf.Width * 2 + (x) * 2)

%assign Pixel.UpDown    0xdb
%assign Pixel.Up        0xdf
%assign Pixel.Down      0xdc
%assign Pixel.Free      0x20
%assign Cursor.Up       0x1e
%assign Cursor.Down     0x1f

;; Keyboard scan codes
;; http://www.delorie.com/djgpp/doc/rbinter/it/06/0.html
%assign Key.ScanCode.Tab      0x0f
%assign Key.ScanCode.Space    0x39
%assign Key.ScanCode.Up       0x48
%assign Key.ScanCode.Down     0x50
%assign Key.ScanCode.Left     0x4b
%assign Key.ScanCode.Right    0x4d
%assign Key.ScanCode.Enter    0x1c
%assign Key.ScanCode.QuitGame 0x01

;; BootLife is supported as both a DOS game and a boot sector game :)

;; Entry point: set up graphics and run game
BootLife:
        cld
        sti                   ;allow interrupts

%ifdef ResetScreen
  ; MDA/VGA text mode 0x03
  ; 80x25 text resolution
        mov ax, 3
        int 0x10

  ; Disable VGA text mode cursor
  ; https://wiki.osdev.org/Text_Mode_Cursor#Disabling_the_Cursor
        mov ah, 0x01
        mov ch, 0x3f
        int 0x10
%endif

;; Run game (the game is restarted by jumping here)
RunGame:
        push cs
        pop ss
        mov sp, TextBuf.Seg

  ; Load VGA text buffer segment into segment registers
        mov es, sp
        push cs
        pop ds

;; 
tick:
        call get_data
data:
%define INITDATA_COLS   36
%define INITDATA_POS    24
%define INITDATA_OFFSET TextBuf.Width*5+10
datainit:
        db                0x01
cursor: db                     0x00
xpos:   db                            0x00,0x00,0x14
ypos:   db                                           0x00
        db 0x00,0x30,0x30
cmd:    db                0x00,0x0c , 0x80,0x08,0x03,0xc0
        db 0x03,0x04,0x31,0x00,0x30 , 0x40,0x34,0x14
cursor_pos:
        db                                           0x00
        db 0x00,0x04,0x01,0x01,0x00 , 0x80,0x08,0x00,0x00
        db 0x00,0x30
datainit_end:
;012345670123456701234567012345670123
;                        X           ;
;                      X X           ;
;            XX      XX            XX;
;           X   X    XX            XX;
;XX        X     X   XX              ;
;XX        X   X XX    X X           ;
;          X     X       X           ;
;           X   X                    ;
;            XX                      ;
;awk 'BEGIN { n = split("0x00,0x00...0x00", t,",")
; i=1; x=strtonum(t[i]); b=8; c=36
; while (i <= n) {
;   if (and(x,1) == 0) printf " "; else printf "X"
;   x=rshift(x,1)
;   if (--c == 0) { printf "\n" ; c=36 ; }
;   if (--b == 0) { x=strtonum(t[++i]) ; b=8 ; }
; }
; printf "\n"
;}' < /dev/null


get_data:
        pop bp
%define BP(x) [bp + x - data]
ZeroBuf:
        mov dx, INITDATA_COLS-INITDATA_POS
        mov cx, TextBuf.Size*2+DigitsNumber+1
        mov bx, BitMap
.1:
        mov byte [bx], dh
        inc bx
        loop .1

        mov bx, BitMap+INITDATA_POS+INITDATA_OFFSET
        mov si, bp
        mov cl, datainit_end-datainit
.next_byte:
        lodsb
        mov ah,8
.next_bit:
        shr al,1
        adc byte [bx], dh
        inc bx
        dec dx
        jnz .noln
        mov dl, INITDATA_COLS
        add bx, TextBuf.Width-INITDATA_COLS
.noln:
        dec ah
        jnz .next_bit
        loop .next_byte

;; Main loop to process key presses and update state
GameLoop:
;; Display board
        xor di, di
        mov si, TextBuf.Width
        mov bx, BitMap
        mov dx, TextBuf.Height + 256
        push bx
.line:
        mov cx, si
.next:
        mov ax, Pixel.Free + 256*Pixel.Up
        test byte [bx + si], dh
        jz .NotDown
        mov ax, Pixel.Down + 256*Pixel.UpDown
.NotDown:
        test byte [bx], dh
        jz .NotUp
        mov al, ah
.NotUp:
        inc bx
        call PutChar
        loop .next
        add bx, si
        dec dl
        jnz .line
        pop bx

        call PutCursor

%ifdef StepCounter
        mov si, BitMap+TextBuf.Size*2+DigitsNumber
.next_digit:
        mov byte [si], ch
        dec si
        inc byte [si]
        cmp byte [si], 10
        je .next_digit
        mov cl, DigitsNumber
        mov si, BitMap+TextBuf.Size*2
        mov di, 6
.digit:
        lodsb
        or al, '0'
        call PutChar
        loop .digit
%endif

        cmp byte BP(cmd), Key.ScanCode.Enter
        jnz wait_key
 
        mov ah,0x01     ; Any key pressed?
        int 0x16
        jz nokey        ; No, go to main loop
wait_key:
        mov ah,0x00
        int 0x16        ; Get key
        mov al, ah
        mov byte BP(cmd), al
        dec ah
        jne notesc      ; No, jump
%ifdef ResetScreen
        mov al,3
        int 0x10
%endif
        int 0x20        ; Exit to DOS or to oblivion (boot sector)
        int 0x19

notesc:
        mov cx, BP(ypos)
        mov di, BP(xpos)

        cmp al, Key.ScanCode.Tab
        jne not_tab
        call GetTextBufIndex
        xor byte [bx + si], 1
GoGameLoop:
        jmp GameLoop

PutCursor:
        xchg ax, dx
        xchg al, BP(cursor)
        mov di, BP(cursor_pos)
        cmp al, 0
        je Ret
PutChar:
        mov [es:di + Vga.TextBuf.Offset], al
        stosb
        inc di
Ret:    ret

not_tab:
CmpUp:
  ; Move cursor up
        dec cx
        cmp al, Key.ScanCode.Up
        je WrapCursor
        inc cx
CmpDown:
  ; Move cursor down
        inc cx
        cmp al, Key.ScanCode.Down
        je WrapCursor
        dec cx
CmpLeft:
  ; Move cursor left
        dec di
        cmp al, Key.ScanCode.Left
        je WrapCursor
        inc di
CmpRight:
  ; Move cursor right
        inc di
        cmp al, Key.ScanCode.Right
        jne nokey
WrapCursor:
        call GetTextBufIndex
        mov BP(ypos), cx
        mov BP(xpos), di
        mov al, Cursor.Up
        shr cx, 1
        adc al, ch              ; Cursor.Down = Cursor.Up + 1
        add cx, cx
        call GetTextBufIndex
        add si, di
        mov BP(cursor_pos), si
        mov BP(cursor), al
GoGameLoop2:
        jmp GoGameLoop

wait_tick:
        hlt                     ; Wait for clock interrupt
nokey:
        mov ah, 0x00            ; Use base clock tick
        int 0x1a
        cmp dx, BP(tick)
        jz wait_tick
        mov BP(tick), dx

;; Compute next board
        mov cx, TextBuf.Height*2-1
.loop_y:
        mov di, TextBuf.Width-1
.loop_x:
        call CountCells         ; si = &TextBuf[cx = y][di = x]
        sub al, 3               ; empty cell with 3 neighbors or full cell with 2 neighbors
        je .alive
        sub al, [bx + si]       ; full cell with 3 neighbors
        jne .dead
.alive:
        or byte [bx + si], 2
.dead:
        dec di
        jns .loop_x
        dec cx
        jns .loop_y

        mov cx, TextBuf.Size*2
.loop_shr:
        shr byte [bx], 1
        inc bx
        loop .loop_shr

        jmp GoGameLoop2

  ; Count adjacent cells
CountCells:
        xor ax, ax

  ; Count left-row
        dec di
        call UpAndCountRow

  ; Count center-row
        call CountNextRow

  ; Count right-row
        call CountNextRow
  
  ; re-center
        dec cx
        dec di

;; Compute the text buffer index from y and x coordinates
;;
;; si = &TextBuf[cx = y][di = x]
;;
;; This computes the equivalent of the TextBuf.Index(y, x) macro, but at runtime
;;
;; Parameters:
;;   * cx - y coordinate
;;   * di - x coordinate
;; Returns:
;;   * si - text buffer index
GetTextBufIndex:
  ; Base case: bounds check y
        mov si, TextBuf.Width
        cmp cx, TextBuf.Height*2
        jb .check_x
        sub cx, TextBuf.Height*2        ; overflow ?
        jns .check_x
        add cx, TextBuf.Height*4        ; underflow !
.check_x:
  ; Base case: bounds check x
        cmp di, si
        jb .check_ok
        sub di, si
        jns .check_ok
        add di, TextBuf.Width*2
.check_ok:
        xchg ax, si
        imul cl
        xchg ax, si
        add si, di
        ret

CountNextRow:
        inc di
        dec cx
UpAndCountRow:
        dec cx
CountRow:
        call Count
        call CountDown
CountDown:
        inc cx
Count:
  ; si = &TextBuf[cx = y][di = x]
        call GetTextBufIndex
%if 0
        cmp [bx + si], ah
%else
        test byte [bx + si], 1
%endif
        je .Ret
        inc ax
.Ret:   ret

;; Print program size at build time
%assign CodeSize $ - $$
%assign Size $ - BootLife
%warning Code is Size bytes

CodeEnd:
  ; Pad to size of boot sector, minus the size of a word for the boot sector
  ; magic value. If the code is too big to fit in a boot sector, the `times`
  ; directive uses a negative value, causing a build error.
        times (BootSector.Size - WordSize) - CodeSize db 0

  ; Boot sector magic
        dw 0xaa55
